@@ -31,7 +31,7 @@ var fs = Promise.promisifyAll(require('fs'));
  * @param {Function} callback Invoked as func(error?, children)
  * @returns A promise
  */
-var getChildren = exports.getChildren = function(fileRoot, workspaceDir, directory, depth, excludes) {
+var getChildren = exports.getChildren = function(fileRoot, workspaceRoot, workspaceDir, directory, depth, excludes) {
 	return fs.readdirAsync(directory)
 	.then(function(files) {
 		return Promise.map(files, function(file) {
@@ -41,15 +41,19 @@ var getChildren = exports.getChildren = function(fileRoot, workspaceDir, directo
 			var filepath = path.join(directory, file);
 			return fs.statAsync(filepath)
 			.then(function(stats) {
-				return fileJSON(fileRoot, workspaceDir, filepath, stats, depth ? depth - 1 : 0);
+				return fileJSON(fileRoot, workspaceRoot, {workspaceDir: workspaceDir, path: filepath}, stats, depth ? depth - 1 : 0);
 			})
 			.catch(function() {
 				return null; // suppress rejection
 			});
 		});
+	}, function reject(err) {
+		return [];
 	})
 	.then(function(results) {
 		return results.filter(function(r) { return r; });
+	}, function reject(err) {
+		return [];
 	});
 };
 
@@ -77,12 +81,40 @@ var safeFilePath = exports.safeFilePath = function(workspaceDir, filepath) {
 	return safePath(workspaceDir, path.join(workspaceDir, filepath));
 };
 
-var getParents = exports.getParents = function(fileRoot, relativePath) {
+var getMetastore = exports.getMetastore = function(req) {
+	var ms = req.app.locals.metastore;
+	if (!ms) {
+		throw new Error("No metastore found");
+	}
+	return ms;
+};
+
+var getFile2 = exports.getFile2 = function(store, rest) {
+	if (rest[0] === "/") rest = rest.substring(1);
+	var segments = rest.split("/");
+	var workspaceId = segments.shift();
+	var workspaceDir = store.getWorkspaceDir(workspaceId);
+	return {
+		workspaceId: workspaceId,
+		workspaceDir: workspaceDir,
+		path: safeFilePath(workspaceDir, segments.join("/"))
+	};
+};
+
+var getFile = exports.getFile = function(req, rest) {
+	if (!rest) return null;
+	var store = getMetastore(req);
+	return getFile2(store, rest);
+};
+
+var getParents = exports.getParents = function(fileRoot, relativePath, includeFolder) {
 	var segs = relativePath.split('/');
 	if(segs && segs.length > 0 && segs[segs.length-1] === ""){// pop the last segment if it is empty. In this case wwwpath ends with "/".
 		segs.pop();
 	}
-	segs.pop();//The last segment now is the directory itself. We do not need it in the parents array.
+	if(!includeFolder) {
+		segs.pop();//The last segment now is the directory itself. We do not need it in the parents array.
+	}
 	var loc = fileRoot;
 	var parents = [];
 	for (var i=0; i < segs.length; i++) {
@@ -98,6 +130,60 @@ var getParents = exports.getParents = function(fileRoot, relativePath) {
 	return parents.reverse();
 };
 
+/**
+ * @description Tries to compute the project path for the given file path
+ * @param {string} workspaceDir The root workspace directory
+ * @param {string} fileRoot The root file path of the server
+ * @param {Object} file The file we want the project for
+ * @param {?} options The optional map of options to use while looking for a project
+ * @returns {Promise} A promise to resolve the project
+ * @since 14.0
+ */
+exports.getProject = function getProject(fileRoot, workspaceRoot, file, options) {
+	var relativePath = file.path;
+	var parents = getParents(fileRoot, relativePath, true),
+		names = options && typeof options.names === "object" ? options.names : {};
+	names['.git'] = {isDirectory: true};
+	names['project.json'] = Object.create(null);
+	if(Array.isArray(parents) && parents.length > 0) {
+		var promises = [];
+		for(var i = 0, len = parents.length; i < len; i++) {
+			var parentFile = parents[i],
+				filepath = parentFile.Location.slice(fileRoot.length);
+			if(filepath.indexOf(file.workspaceDir) !== 0) {
+				break;
+			}
+			promises.push(findProject(fileRoot, workspaceRoot, file.workspaceDir, filepath, names));
+		}
+		return Promise.any(promises);
+	}
+	return Promise.resolve(null);
+};
+/**
+ * @description Finds the project from the given file context
+ * @param {string} fileRoot The root file context
+ * @param {string} workspaceDir The root workspace location
+ * @param {string} filepath The absolute path to the file context we are starting from
+ * @param {?} names The map of names to check against to determine project-ness
+ * @since 14.0
+ * @returns {Promise} Returns a promise to try and resolve the project from the context
+ */
+function findProject(fileRoot, workspaceRoot, workspaceDir, filepath, names) {
+	return new Promise(function(resolve, reject) {
+		return getChildren(fileRoot, workspaceRoot, workspaceDir, filepath, 1).then(function(children) {
+			if(Array.isArray(children) && children.length > 0) {
+				for(var i = 0, len = children.length; i < len; i++) {
+					var c = children[i],
+						n = names[c.Name];
+					if(n && (c.Directory && n.isDirectory || !c.Directory && !n.isDirectory)) {
+						return resolve(filepath);
+					}
+				}
+			}
+			return reject(new Error('not a project'));
+		});
+	});
+}
 /**
  * Performs the equivalent of rm -rf on a directory.
  * @param {Function} callback Invoked as callback(error)
@@ -182,7 +268,11 @@ exports.withETag = function(filepath, callback) {
  * @returns {String} The Location of for a file resource.
  */
 function getFileLocation(fileRoot, wwwpath, isDir) {
-	return api.join(fileRoot, wwwpath) + (isDir ? '/' : '');
+	var filepath = api.join(fileRoot, wwwpath);
+	if(isDir && filepath.lastIndexOf('/') !== filepath.length-1) {
+		filepath += '/';
+	}
+	return filepath;
 }
 
 /**
@@ -197,6 +287,17 @@ function getBoolean(obj, key) {
 	return Object.prototype.hasOwnProperty.call(obj, key) && (val === true || val === 'true');
 }
 
+/**
+ * File decorator interface.
+ */
+exports.FileDecorator = FileDecorator;
+function FileDecorator() {
+}
+Object.assign(FileDecorator.prototype, {
+	decorate: function(req, file, json) {
+	}
+});
+
 var decorators = [];
 /**
  * Shared decorator, used by workspace as well.
@@ -206,30 +307,31 @@ exports.getDecorators = function(){
 };
 /**
  * Used to add different decorators to generate respond json.
- * @param {func} decorator functions to be added;
+ * @param {FileDecorator} decorator to be added;
  */
-exports.addDecorator = function(func) {
-	decorators.push(func);
+exports.addDecorator = function(decorator) {
+	decorators.push(decorator);
 };
 
 /**
  * Helper for fulfilling a file metadata GET request.
  * @param {String} fileRoot The "/file" prefix or equivalent.
+ * @param {String} workspaceRoot The "/file" prefix or equivalent.
  * @param {Object} req HTTP request object
  * @param {Object} res HTTP response object
- * @param {String} filepath The physical path to the file on the server.
+ * @param {String} file The physical path and workspaceDir to the file on the server.
  * @param {Object} stats
  * @param {String} etag
  * @param {Boolean} [includeChildren=false]
  * @param {Object} [metadataMixins] Additional metadata to mix in to the response object.
  */
-var writeFileMetadata = exports.writeFileMetadata = function(fileRoot, req, res, contextPath, filepath, stats, etag, depth, metadataMixins) {
+var writeFileMetadata = exports.writeFileMetadata = function(req, res, fileRoot, workspaceRoot, file, stats, etag, depth, metadataMixins) {
 	var result;
-	return fileJSON(fileRoot, req.user.workspaceDir, filepath, stats, depth, metadataMixins)
+	return fileJSON(fileRoot, workspaceRoot, file, stats, depth, metadataMixins)
 	.then(function(originalJson) {
 		result = originalJson;
 		return Promise.map(decorators, function(decorator){
-			return decorator(contextPath, fileRoot, req, filepath, result);			
+			return decorator.decorate(req, file, result);			
 		});
 	})
 	.then(function(){
@@ -237,17 +339,16 @@ var writeFileMetadata = exports.writeFileMetadata = function(fileRoot, req, res,
 			result.ETag = etag;
 			res.setHeader('ETag', etag);
 		}
-		res.setHeader("Cache-Control", "no-cache");
-		api.write(null, res, null, result);
+		api.writeResponse(null, res, null, result, true, true);
 	})
 	.catch(api.writeError.bind(null, 500, res));
 };
-function fileJSON(fileRoot, workspaceDir, filepath, stats, depth, metadataMixins) {
+function fileJSON(fileRoot, workspaceRoot, file, stats, depth, metadataMixins) {
 	depth = depth || 0;
 	var isDir = stats.isDirectory();
-	var wwwpath = api.toURLPath(filepath.substring(workspaceDir.length + 1));
+	var wwwpath = api.toURLPath(file.path.substring(file.workspaceDir.length + 1));
 	var result = {
-		Name: path.basename(filepath),
+		Name: path.basename(file.path),
 		Location: getFileLocation(fileRoot, wwwpath, isDir),
 		Directory: isDir,
 		LocalTimeStamp: stats.mtime.getTime(),
@@ -263,6 +364,7 @@ function fileJSON(fileRoot, workspaceDir, filepath, stats, depth, metadataMixins
 			result[property] = metadataMixins[property];
 		});
 	}
+	result.WorkspaceLocation = workspaceRoot;
 	if (!isDir) {
 		return Promise.resolve(result);
 	}
@@ -273,7 +375,7 @@ function fileJSON(fileRoot, workspaceDir, filepath, stats, depth, metadataMixins
 	if (depth <= 0) {
 		return Promise.resolve(result);
 	}
-	return getChildren(fileRoot, workspaceDir, filepath, depth)
+	return getChildren(fileRoot, workspaceRoot, file.workspaceDir, file.path, depth)
 	.then(function(children) {
 		result.Children = children;
 		return result;
@@ -283,15 +385,16 @@ function fileJSON(fileRoot, workspaceDir, filepath, stats, depth, metadataMixins
 /**
  * Helper for fulfilling a file POST request (for example, copy, move, or create).
  * @param {String} fileRoot The route of the /file handler (not including context path)
+ * @param {String} workspaceRoot The route of the /workspace handler (not including context path)
  * @param {Object} req
  * @parma {Object} res
  * @param {String} wwwpath
- * @param {String} destFilepath
+ * @param {String} destFile
  * @param {Object} [metadata] Additional metadata to be mixed in to the File response.
  * @param {Number} [statusCode] Status code to send on a successful response. By default, `201 Created` is sent if
  * a new resource was created, and and `200 OK` if an existing resource was overwritten.
  */
-exports.handleFilePOST = function(contextPath, fileRoot, req, res, destFilepath, metadataMixins, statusCode) {
+exports.handleFilePOST = function(workspaceRoot, fileRoot, req, res, destFile, metadataMixins, statusCode) {
 	var isDirectory = req.body && getBoolean(req.body, 'Directory');
 	var writeResponse = function(isOverwrite) {
 		// TODO: maybe set ReadOnly and Executable based on fileAtts
@@ -301,15 +404,15 @@ exports.handleFilePOST = function(contextPath, fileRoot, req, res, destFilepath,
 			// Status code 200 indicates that an existing resource was replaced, or we're POSTing to a URL
 			res.status(isOverwrite ? 200 : 201);
 		}
-		return fs.statAsync(destFilepath)
+		return fs.statAsync(destFile.path)
 		.then(function(stats) {
-			return writeFileMetadata(fileRoot, req, res, contextPath, destFilepath, stats, /*etag*/null, /*depth*/0, metadataMixins);
+			return writeFileMetadata(req, res, api.join(fileRoot, destFile.workspaceId), api.join(workspaceRoot, destFile.workspaceId), destFile, stats, /*etag*/null, /*depth*/0, metadataMixins);
 		})
 		.catch(function(err) {
 			api.writeError(500, res, err.message);
 		});
 	};
-	fs.statAsync(destFilepath)
+	fs.statAsync(destFile.path)
 	.catchReturn({ code: 'ENOENT' }, null) // suppress reject when file does not exist
 	.then(function(stats) {
 		return !!stats; // just return whether the file existed
@@ -318,7 +421,7 @@ exports.handleFilePOST = function(contextPath, fileRoot, req, res, destFilepath,
 		var xCreateOptions = (req.headers['x-create-options'] || "").split(",");
 		var isCopy = xCreateOptions.indexOf('copy') !== -1, isMove = xCreateOptions.indexOf('move') !== -1;
 		if (isCopy && isMove) {
-			return api.write(400, res, null, 'Illegal combination of X-Create-Options.');
+			return api.writeResponse(400, res, null, 'Illegal combination of X-Create-Options.', true);
 		}
 		if (xCreateOptions.indexOf('no-overwrite') !== -1 && destExists) {
 			return api.writeError(412, res, new Error('A file or folder with the same name already exists at this location.'));
@@ -329,10 +432,10 @@ exports.handleFilePOST = function(contextPath, fileRoot, req, res, destFilepath,
 			if (!sourceUrl) {
 				return api.writeError(400, res, 'Missing Location property in request body');
 			}
-			var sourceFilepath = safeFilePath(req.user.workspaceDir, api.rest(fileRoot, api.matchHost(req, sourceUrl)));
-			return fs.statAsync(sourceFilepath)
+			var sourceFile = getFile(req, sourceUrl.replace(/^\/file/, ""));
+			return fs.statAsync(sourceFile.path)
 			.then(function(/*stats*/) {
-				return isCopy ? copy(sourceFilepath, destFilepath) : fs.renameAsync(sourceFilepath, destFilepath);
+				return isCopy ? copy(sourceFile.path, destFile.path) : fs.renameAsync(sourceFile.path, destFile.path);
 			})
 			.then(writeResponse.bind(null, destExists))
 			.catch(function(err) {
@@ -345,10 +448,10 @@ exports.handleFilePOST = function(contextPath, fileRoot, req, res, destFilepath,
 		// Just a regular file write
 		return Promise.resolve()
 		.then(function(){
-			return destExists ? fs.unlinkAsync(destFilepath) : null;
+			return destExists ? fs.unlinkAsync(destFile.path) : null;
 		})
 		.then(function() {
-			return isDirectory ? fs.mkdirAsync(destFilepath) : fs.writeFileAsync(destFilepath, '');
+			return isDirectory ? fs.mkdirAsync(destFile.path) : fs.writeFileAsync(destFile.path, '');
 		})
 		.then(writeResponse.bind(null, destExists))
 		.catch(api.writeError.bind(null, 500, res));
