@@ -479,9 +479,6 @@ define([
 		for (var i=0; i<links.length; i++) {
 			var link = links[i];
 			link.addEventListener("click", clicked, false);
-			if(util.isElectron){
-				link.target = "_blank";
-			}
 		}
 	};
 	/** @private */
